/*====================================================================
		rblob.c

		rFunc InterBase UDF library.
		Blob functions.

		Copyright 1998-2001 Polaris Software
		http://polesoft.da.ru
		mailto: rFunc@mail.ru

		Version 2.0.1.2
====================================================================== */

#include <ibase.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "rfunc.h"
#include "rblob.h"

long EXPORT fn_b_number_segments (ARG(BLOB, b))
ARGLIST(BLOB b)
{
	if (!b->blob_handle)
		return 0L;

	return b->blob_number_segments;
}

long EXPORT fn_b_max_segment (ARG(BLOB, b))
ARGLIST(BLOB b)
{
	if (!b->blob_handle)
		return 0L;

	return b->blob_max_segment;
}

long EXPORT fn_b_total_length (ARG(BLOB, b))
ARGLIST(BLOB b)
{
	if (!b->blob_handle)
		return 0L;

	return b->blob_total_length;
}

long EXPORT fn_b_line_count (ARG(BLOB, b))
ARGLIST(BLOB b)
{
	char *buf, *p;
	long curr_segment = 0L;
	unsigned short length, actual_length;
	long linecount = 0L;

	if (!b->blob_handle)
		return 0L;

	length = b->blob_max_segment + 1L;
	buf = (char *) malloc (b->blob_max_segment + 1L);

	while ((*b->blob_get_segment) (b->blob_handle, buf, length, &actual_length))
	{
		buf [actual_length] = 0;
		p = buf;
		while (*p)
			if (*p++ == '\n')
				linecount++;
		if ((++curr_segment == b->blob_number_segments) && actual_length && (*--p != '\n'))
			linecount++;
	}

	free (buf);
	return linecount;
}

char* b_nsubstr(ARG(BLOB, b), ARG(long*, m), ARG(long*, n), ARG(long, maxlength))
ARGLIST(BLOB b)
ARGLIST(long *m)
ARGLIST(long *n)
ARGLIST(long maxlength)
{
	char *buf, *p, *q;
//  long i = 0;
	long curr_bytecount = 0;
	unsigned short length, actual_length;

	char *buffer = (char *) MALLOC (maxlength);

	long left, right;

	left = (*m < 0) ? b->blob_total_length + *m + 1 : *m;
	right = (*n < 0) ? left - 1 : left - 1 + *n - 1;
	left = (*n < 0) ? left + *n : left - 1;

	left = (left < 0) ? 0 : left;
	right = (right > b->blob_total_length - 1) ? b->blob_total_length - 1 : right;

	if (!*n || (left > b->blob_total_length - 1) || (right < 0))
		return "";

	if (!b->blob_handle)
		return "";

	left = (left < 0) ? 0 : left;
	right = (right > b->blob_total_length - 1) ? b->blob_total_length - 1 : right;

	if (!*n || (left > b->blob_total_length - 1) || (right < 0)) return "";

	length = b->blob_max_segment + 1L;
	buf = (char *) malloc (length);

	// Limit the return string to "maxlength" bytes
	if (right - left + 1L > maxlength - 1L)
		right = left + maxlength - 2L;
	q = buffer;

	while ((*b->blob_get_segment) (b->blob_handle, buf, length,
		&actual_length))
	{
		buf [actual_length] = 0;

		p = buf;
		while (*p && (curr_bytecount <= right))
		{
//      if (*p == '\n')
//        *p = ' ';
			if (curr_bytecount >= left)
				*q++ = *p;
			p++;
			curr_bytecount++;
		}
		if (curr_bytecount >= right)
		{
			*q = 0;
			break;
		}
	}

	free (buf);
	return buffer;
}

char* EXPORT fn_b_substr(ARG(BLOB, b), ARG(long*, m), ARG(long*, n))
ARGLIST(BLOB b)
ARGLIST(long *m)
ARGLIST(long *n)
{
	return b_nsubstr(b, m, n, shortlen);
}

char* EXPORT fn_b_longsubstr(ARG(BLOB, b), ARG(long*, m), ARG(long*, n))
ARGLIST(BLOB b)
ARGLIST(long *m)
ARGLIST(long *n)
{
	return b_nsubstr(b, m, n, longlen);
}

char* b_nline(ARG(BLOB, b), ARG(long*, l), ARG(long, maxlength))
ARGLIST(BLOB b)
ARGLIST(long *l)
ARGLIST(long maxlength)
{
	char *buf, *p;
//  long i = 0;
	long curr_bytecount = 0, curr_linecount = 1;
	unsigned short length, actual_length;
	char *buffer = (char *) MALLOC (maxlength);

	buffer[0] = 0;
	if (!b->blob_handle)
		return buffer;

	length = b->blob_max_segment + 1L;
	buf = (char *) malloc (length);
	while ((*b->blob_get_segment) (b->blob_handle, buf, length,
		&actual_length))
	{
		buf [actual_length] = 0;

		p = buf;
		while (*p)
		{
//      if ((*p == 13) || (*p == 10))
			if (*p == '\n')
			{
				curr_linecount++;
				if (curr_linecount > *l)
					break;
			}
			else
			if (curr_linecount == *l)
			{
				if (*p != 10)
					buffer[curr_bytecount++] = *p;
				if (curr_bytecount >= maxlength - 1L)
					break;
			}
			p++;
		}
		if ((curr_linecount > *l) || (curr_bytecount >= maxlength - 1L))
			break;
	}

	buffer[curr_bytecount] = 0;
	free (buf);
	return buffer;
}

char* EXPORT fn_b_line(ARG(BLOB, b), ARG(long*, l))
ARGLIST(BLOB b)
ARGLIST(long *l)
{
	return b_nline(b, l, shortlen);
}

char* EXPORT fn_b_longline(ARG(BLOB, b), ARG(long*, l))
ARGLIST(BLOB b)
ARGLIST(long *l)
{
	return b_nline(b, l, longlen);
}

void EXPORT fn_b_put_segment(ARG(char*, s), ARG(BLOB, b))
ARGLIST(char* s)
ARGLIST(BLOB b)
{
	if (b->blob_handle)
		(*b->blob_put_segment) (b->blob_handle, s, strlen(s));
}
